from enum import Enum
from dataclasses import dataclass


@dataclass
class Subsystem:
    """
    Defines constants for the available subsystems
    """
    description: str


class Subsystems(Enum):
    AOESW = Subsystem("Adaptive Optics Executive Software")
    APS = Subsystem("Alignment and Phasing System ")
    CIS = Subsystem("Communications and Information Systems")
    CLN = Subsystem("Optical Cleaning Systems")
    CRYO = Subsystem("Instrumentation Cryogenic Cooling System")
    CSW = Subsystem("Common Software")
    DMS = Subsystem("Data Management System")
    DPS = Subsystem("Data Processing System")
    ENC = Subsystem("Enclosure")
    ESEN = Subsystem("Engineering Sensors")
    ESW = Subsystem("Executive Software")
    HNDL = Subsystem("Optics Handling Equipment")
    HQ = Subsystem("Observatory Headquarters")
    IRIS = Subsystem("InfraRed Imaging Spectrometer")
    LGSF = Subsystem("Laser Guide Star Facility")
    M1COAT = Subsystem("M1COAT M1 Optical Coating System")
    M1CS = Subsystem("M1CS M1 Control System ")
    M1S = Subsystem("M1S M1 Optics System")
    M2COAT = Subsystem("M2/M3 Optical Coating System")
    M2S = Subsystem("M2S M2 System")
    M3S = Subsystem("M3S M3 System")
    MODHIS = Subsystem("Multi-Object Diffraction-limited High-resolution IR Spectrograph")
    NFIRAOS = Subsystem("Narrow Field Infrared AO System")
    OSS = Subsystem("Observatory Safety System")
    REFR = Subsystem("Instrumentation Refrigerant Cooling System ")
    SCMS = Subsystem("Site Conditions Monitoring System")
    SER = Subsystem("Services")
    SOSS = Subsystem("Science Operations Support Systems")
    STR = Subsystem("Structure ")
    SUM = Subsystem("Summit Facilities")
    TCS = Subsystem("Telescope Control System")
    TINS = Subsystem("Test Instruments")
    WFOS = Subsystem("Wide Field Optical Spectrograph")
    Container = Subsystem("Container Subsystem")



from enum import Enum


class KeyType(Enum):
    ChoiceKey = 1
    StringKey = 2
    UTCTimeKey = 3
    TAITimeKey = 4
    EqCoordKey = 5
    SolarSystemCoordKey = 6
    MinorPlanetCoordKey = 7
    CometCoordKey = 8
    AltAzCoordKey = 9
    CoordKey = 10
    BooleanKey = 11
    CharKey = 12
    ByteKey = 13
    ShortKey = 14
    LongKey = 15
    IntKey = 16
    FloatKey = 17
    DoubleKey = 18
    ByteArrayKey = 19
    ShortArrayKey = 20
    LongArrayKey = 21
    IntArrayKey = 22
    FloatArrayKey = 23
    DoubleArrayKey = 24
    ByteMatrixKey = 25
    ShortMatrixKey = 26
    LongMatrixKey = 27
    IntMatrixKey = 28
    FloatMatrixKey = 29
    DoubleMatrixKey = 30



from dataclasses import dataclass
from typing import List

from csw.Parameter import Parameter


@dataclass
class CommandResponse:
    """
    Type of a response to a command (submit, oneway or validate).
    Note that oneway and validate responses are limited to Accepted, Invalid or Locked.
    """
    runId: str

    def _asDict(self):
        """
        Returns: XXX: a dictionary corresponding to this object
        """
        return {
            "_type": self.__class__.__name__,
            'runId': self.runId,
        }


@dataclass
class Cancelled(CommandResponse):
    """Represents a negative response that describes the cancellation of command"""
    pass


@dataclass
class Accepted(CommandResponse):
    """Represents a final response stating acceptance of a command received"""
    pass


@dataclass
class Error(CommandResponse):
    """Represents a negative response that describes an error in executing the command"""
    message: str

    def _asDict(self):
        """
        Returns: dict
            a dictionary corresponding to this object
        """
        return {
            "_type": self.__class__.__name__,
            'runId': self.runId,
            'message': self.message
        }


@dataclass
class Locked(CommandResponse):
    """Represents a negative response stating that a component is Locked and command was not validated or executed"""
    message: str

    def _asDict(self):
        """
        Returns: dict
            a dictionary corresponding to this object
        """
        return {
            "_type": self.__class__.__name__,
            'runId': self.runId,
            'message': self.message
        }


@dataclass
class Started(CommandResponse):
    """Represents an intermediate response stating a long running command has been started"""
    message: str

    def _asDict(self):
        """
        Returns: dict
            a dictionary corresponding to this object
        """
        return {
            "_type": self.__class__.__name__,
            'runId': self.runId,
            'message': self.message
        }


@dataclass
class Result:
    """A result containing parameters for command response"""
    paramSet: List[Parameter]

    # noinspection PyProtectedMember
    def _asDict(self):
        """
        Returns: dict
            a dictionary corresponding to this object
        """
        return {
            'paramSet': list(map(lambda p: p._asDict(), self.paramSet))
        }


@dataclass
class Completed(CommandResponse):
    """Represents a positive response stating completion of command"""
    result: Result = Result([])

    # noinspection PyProtectedMember
    def _asDict(self):
        """
        Returns: dict
            a dictionary corresponding to this object
        """
        return {
            "_type": self.__class__.__name__,
            'runId': self.runId,
            'result': self.result._asDict()
        }


# --- Invalid ---
@dataclass
class CommandIssue:
    """Describes a command issue with appropriate reason for validation failure"""
    reason: str


class IdNotAvailableIssue(CommandIssue):
    """Returned when a CommandResponse associated with runId is not available"""


class HCDBusyIssue(CommandIssue):
    """Returned when the HCD is busy and can't process a command"""


class WrongCommandTypeIssue(CommandIssue):
    """Returned when some given command type is not expected"""


class MissingKeyIssue(CommandIssue):
    """Returned when a command is missing a required key/parameter"""


class WrongPrefixIssue(CommandIssue):
    """Returned when an Assembly receives a configuration with a prefix that it doesn't support"""


class WrongParameterTypeIssue(CommandIssue):
    """Returned when the parameter for a key is not the correct type (i.e. int vs double, etc.)"""


class WrongUnitsIssue(CommandIssue):
    """Returned when a parameter value does not have the correct units"""


class WrongNumberOfParametersIssue(CommandIssue):
    """Returned when a command does not have the correct number of parameters"""


class AssemblyBusyIssue(CommandIssue):
    """Returned when an Assembly receives a command and one is already executing"""


class UnresolvedLocationsIssue(CommandIssue):
    """Returned when some required location is not available"""


class ParameterValueOutOfRangeIssue(CommandIssue):
    """Parameter of a command is out of range"""


class WrongInternalStateIssue(CommandIssue):
    """The component is in the wrong internal state to handle a command"""


class UnsupportedCommandInStateIssue(CommandIssue):
    """A command is unsupported in the current state"""


class UnsupportedCommandIssue(CommandIssue):
    """A command is unsupported by component"""


class RequiredServiceUnavailableIssue(CommandIssue):
    """A required service is not available"""


class RequiredHCDUnavailableIssue(CommandIssue):
    """A required HCD is not available"""


class RequiredAssemblyUnavailableIssue(CommandIssue):
    """A required Assembly is not available"""


class RequiredSequencerUnavailableIssue(CommandIssue):
    """Returned when some other issue occurred apart from those already defined"""


class OtherIssue(CommandIssue):
    """A required Sequencer is not available"""


@dataclass
class Invalid(CommandResponse):
    issue: CommandIssue

    def _asDict(self):
        """
        Returns: dict
            a dictionary for this object
        """
        return {
            "_type": self.__class__.__name__,
            'runId': self.runId,
            'issue': {
                "_type": self.issue.__class__.__name__,
                "reason": self.issue.reason
            }
        }



import cbor2

from csw.Event import Event
from csw.RedisConnector import RedisConnector


class EventPublisher:

    def __init__(self):
        self.__redis = RedisConnector()

    def publish(self, event: Event):
        """
        Publish an event to the Event Service

        Args:
            event (Event): Event to be published
        """
        event_key = str(event.source) + "." + event.eventName.name
        obj = cbor2.dumps(event._asDict())
        self.__redis.publish(event_key, obj)



import asyncio
from asyncio import Task

from csw.CommandResponse import CommandResponse, Error


class CommandResponseManager:
    """
    Keeps track of command responses for long running commands by maintaining a
    map from runId to CommandResponse.
    """
    tasks = {}

    def addTask(self, runId: str, task: Task):
        self.tasks[runId] = task

    async def waitForTask(self, runId: str, timeoutInSeconds: int) -> CommandResponse:
        if runId in self.tasks:
            task: Task = self.tasks[runId]
            return await asyncio.wait_for(task, timeout=timeoutInSeconds*1.0)
        else:
            return Error(runId, "No task was found for runId " + runId)



from enum import Enum
from dataclasses import dataclass


@dataclass
class Unit:
    """
    A class representing units for TMT (Corresponds to the CSW Units Scala class)
    """
    name: str
    description: str


class Units(Enum):
    # SI units
    angstrom = Unit("Angstrom", "10 -1 nm")
    arcmin = Unit("arcmin", "arc minute; angular measurement")
    arcsec = Unit("arcsec", "arc second: angular measurement")
    day = Unit("d", "day - 24 hours")
    degree = Unit("deg", "degree: agular measurement 1/360 of full rotation")
    elvolt = Unit("eV", "electron volt 1.6022x10-19 J")
    gram = Unit("g", "gram 10-3 kg")
    hour = Unit("h", "hour 3.6x10+3 s")
    hertz = Unit("Hz", "frequency")
    joule = Unit("J", "Joule: energy N m")
    kelvin = Unit("K", "Kelvin: temperature with a null point at absolute zero")
    kilogram = Unit("kg", "kilogram, base unit of mass in SI")
    kilometer = Unit("km", "kilometers - 10+3 m")
    liter = Unit("l", "liter, metric unit of volume 10+3 cm+3")
    meter = Unit("m", "meter: base unit of length in SI")
    marcsec = Unit("mas", "milli arc second: angular measurement 10-3 arcsec")
    millimeter = Unit("mm", "millimeters - 10-3 m")
    millisecond = Unit("ms", "milliseconds - 10-3 s")
    micron = Unit("µm", "micron: alias for micrometer")
    micrometer = Unit("µm", "micron: 10-6 m")
    minute = Unit("min", "minute 6x10+1 s")
    newton = Unit("N", "Newton: force")
    pascal = Unit("Pa", "Pascal: pressure")
    radian = Unit("rad", "radian: angular measurement of the ratio between the length of an arc and its radius")
    second = Unit("s", "second: base unit of time in SI")
    sday = Unit("sday", "sidereal day is the time of one rotation of the Earth: 8.6164x10+4 s")
    steradian = Unit("sr", "steradian: unit of solid angle in SI - rad+2")
    microarcsec = Unit("µas", "micro arcsec: angular measurement")
    volt = Unit("V", "Volt: electric potential or electromotive force")
    watt = Unit("W", "Watt: power")
    week = Unit("wk", "week - 7 d")
    year = Unit("yr", "year - 3.6525x10+2 d")
    # CGS units
    coulomb = Unit("C", "coulomb: electric charge")
    centimeter = Unit("cm", "centimeter")
    erg = Unit("erg", "erg: CGS unit of energy")
    # Astropyhsics units
    au = Unit("AU", "astronomical unit: approximately the mean Earth-Sun distance")
    jansky = Unit("Jy", "Jansky: spectral flux density - 10-26 W/Hz m+2")
    lightyear = Unit("lyr", "light year - 9.4607x10+15 m")
    mag = Unit("mag", "stellar magnitude")
    # Imperial units
    cal = Unit("cal", "thermochemical calorie: pre-SI metric unit of energy")
    foot = Unit("ft", "international foot - 1.2x10+1 inch")
    inch = Unit("inch", "international inch - 2.54 cm")
    pound = Unit("lb", "international avoirdupois pound - 1.6x10+1 oz")
    mile = Unit("mi", "international mile - 5.28x10+3 ft")
    ounce = Unit("oz", "international avoirdupois ounce")
    yard = Unit("yd", "international yard - 3 ft")
    # Others - engineering
    NoUnits = Unit("none", "scalar - no units specified")
    encoder = Unit("enc", "encoder counts")
    count = Unit("ct", "counts as for an encoder or detector")
    pix = Unit("pix", "pixel")
    # Datetime units
    tai = Unit("TAI", "TAI time unit")
    utc = Unit("UTC", "UTC time unit")



from urllib.parse import urlparse

import redis
from typing import List

from redis.sentinel import Sentinel

from csw.LocationService import ConnectionInfo, ComponentType, ConnectionType, LocationService
from csw.Prefix import Prefix
from csw.Subsystem import Subsystems


class RedisConnector:

    def __init__(self):
        """
        Events are posted to Redis. This is internal class used to access Redis.
        """
        prefix = Prefix(Subsystems.CSW, "EventServer")
        conn = ConnectionInfo.make(prefix, ComponentType.Service, ConnectionType.TcpType)
        loc = LocationService().find(conn)
        uri = urlparse(loc.uri)
        sentinel = Sentinel([(uri.hostname, uri.port)], socket_timeout=0.1)
        self.__redis = sentinel.master_for('eventServer', socket_timeout=0.1)
        self.__redis_pubsub = self.__redis.pubsub()

    def close(self):
        self.__redis_pubsub.close()

    def subscribe(self, keyList: List[str], callback):
        """
        Set up a Redis subscription on specified keys with specified callback on value changes.

        Args:
            keyList (List[str]): list of keys to subscribe to
            callback (function): callback called when item changes.  Should take a Redis message type.

        Returns: PubSubWorkerThread
            subscription thread. Use .stop() method to stop subscription
        """
        d = dict.fromkeys(keyList, callback)
        self.__redis_pubsub.subscribe(**d)
        return self.__redis_pubsub.run_in_thread(sleep_time=0.001)

    def unsubscribe(self, keyList: List[str]):
        """
        Unsubscribe to the list of event keys

        Args:
            keyList (List[str]): list of keys to unsubscribe from
        """
        self.__redis_pubsub.unsubscribe(keyList)

    # XXX Commented out due to Event Service performance concerns when using psubscribe
    # def pSubscribe(self, keyList: List[str], callback):
    #     """
    #     Set up a Redis subscription on specified keys with specified callback on value changes.
    #     In this case the keys are treated as glob-style patterns.
    #
    #     Args:
    #         keyList (List[str]): list of key patterns to subscribe to
    #         callback (function): callback called when item changes.  Should take a Redis message type.
    #
    #     Returns: PubSubWorkerThread
    #         subscription thread. Use .stop() method to stop subscription
    #     """
    #     d = dict.fromkeys(keyList, callback)
    #     self.__redis_pubsub.psubscribe(**d)
    #     return self.__redis_pubsub.run_in_thread(sleep_time=0.001)

    # def pUnsubscribe(self, keyList: List[str]):
    #     """
    #     Unsubscribe to the list of event key patterns
    #
    #     Args:
    #         keyList (List[str]): list of key patterns to unsubscribe from
    #     """
    #     self.__redis_pubsub.punsubscribe(keyList)

    def publish(self, key: str, encodedValue: bytes):
        """
        Publish CBOR encoded event string to Redis

        Args:
            key: String specifying Redis key for event.  Should be source prefix + "." + event name.
            encodedValue: CBOR encoded value for the event (in the form [className, dict])
        """
        self.__redis.set(key, encodedValue)
        self.__redis.publish(key, encodedValue)

    def get(self, key: str) -> str:
        """
        Get value from Redis using specified key

        Args:
            key (str): String specifying Redis key for event.  Should be source prefix + "." + event name.

        Returns: str
            Raw Redis string for event, typically in some encoding
        """
        return self.__redis.get(key)



from dataclasses import dataclass

from csw.TAITime import TAITime
from csw.UTCTime import UTCTime
from csw.Units import Units
from csw.KeyType import KeyType
from csw.Coords import Coord

coordTypes = {KeyType.CoordKey, KeyType.EqCoordKey, KeyType.SolarSystemCoordKey, KeyType.MinorPlanetCoordKey,
              KeyType.CometCoordKey}
timeKeyTypes = {KeyType.TAITimeKey, KeyType.UTCTimeKey}


@dataclass
class Parameter:
    """
    Creates a Parameter (keys with values, units).

    Args:

        keyName (str): name of the key
        keyType (KeyType): type of the key
        values (object): an array of values, or a nested array for array and matrix types.
        units (Units): units of the values.
    """
    keyName: str
    keyType: KeyType
    values: object
    units: Units = Units.NoUnits

    @staticmethod
    def _paramValueOrDict(keyType: KeyType, param, forEvent: bool):
        """
        Internal method that also handles coord and time types

        Args:
            keyType (KeyType): parameter's key type
            param: parameter value
            forEvent (bool): needed since time values are encoded differently in CBOR and JSON

        Returns:
            param value
        """
        # keyTypes = coordTypes.union(timeKeyTypes) if forEvent else coordTypes
        if keyType in coordTypes:
            return param._asDict()
        if keyType in timeKeyTypes:
            if forEvent:
                return param._asDict()
            return str(param)
        return param

    @staticmethod
    def _paramValueFromDict(keyType: KeyType, obj, forEvent: bool):
        """
        Internal recursive method that also handles Coord types

        Args:

            keyType (KeyType): parameter's key type
            obj: parameter value
            forEvent (bool): needed since time values are encoded differently in CBOR and JSON

        Returns: object
            param value
        """
        if keyType in coordTypes:
            return Coord._fromDict(obj)
        elif not forEvent and keyType in timeKeyTypes:
            if keyType == KeyType.UTCTimeKey:
                return UTCTime.from_str(obj)
            else:
                return TAITime.from_str(obj)
        else:
            return obj

    # noinspection PyTypeChecker
    # forEvent flag is needed since time values are encoded differently in CBOR and JSON
    def _asDict(self, forEvent: bool = False):
        # Note that bytes are stored in a byte string (b'...') instead of a list or array.
        if self.keyType == KeyType.ByteKey:
            values = self.values
        else:
            values = list(map(lambda p: Parameter._paramValueOrDict(self.keyType, p, forEvent), self.values))

        return {
            self.keyType.name: {
                'keyName': self.keyName,
                'values': values,
                'units': self.units.name
            }
        }

    @staticmethod
    # forEvent flag is needed since time values are encoded differently in CBOR and JSON
    def _fromDict(obj: dict, forEvent: bool = False):
        """
        Returns a Parameter for the given dict.
        """
        k = next(iter(obj))
        keyType = KeyType[k]
        obj = obj[k]

        if keyType == KeyType.ByteKey:
            values = obj['values']
        else:
            values = list(map(lambda p: Parameter._paramValueFromDict(keyType, p, forEvent), obj['values']))
        return Parameter(obj['keyName'], keyType, values, Units[obj['units']])



from dataclasses import dataclass

from csw.Subsystem import Subsystems


@dataclass
class Prefix:
    """
    A class representing a component prefix, made up of the subsystem and the component name.
    Note: Component name should not contain leading or trailing spaces or a hyphen (-).
    """
    subsystem: Subsystems
    componentName: str

    def __init__(self, subsystem: Subsystems, componentName: str):
        assert (componentName == componentName.strip())
        assert ("-" not in componentName)
        self.subsystem = subsystem
        self.componentName = componentName

    def __str__(self):
        return f"{self.subsystem.name}.{self.componentName}"

    @classmethod
    def from_str(class_object, prefixStr: str):
        [s, c] = prefixStr.split('.', 1)
        return class_object(Subsystems[s.upper()], c)



import uuid
from dataclasses import dataclass
from typing import List
from abc import abstractmethod

from csw.Parameter import Parameter
from csw.EventTime import EventTime
from csw.Prefix import Prefix

@dataclass
class EventName:
    """
    A wrapper class representing the name of an Event
    """
    name: str

@dataclass
class Event:
    """
    Abstract base class that creates an Event that can be published to the event service
    (Don't use this class directly: The system expects a SystemEvent or an ObserveEvent).

    Args:
        source (Prefix): prefix representing source of the event
        eventName (EventName): the name of event
        paramSet (list): list of Parameter (keys with values)
        eventTime (EventTime): the time the event was created (defaults to current time)
        eventId (str): event id (optional: Should leave empty unless received from event service)
    """
    source: Prefix
    eventName: EventName
    paramSet: List[Parameter]
    eventTime: EventTime = EventTime.fromSystem()
    eventId: str = str(uuid.uuid4())

    @abstractmethod
    def eventType(self) -> str:
        """
        This is only here to make the Event class abstract.

        Returns: str
            the type of the event
        """
        pass

    @staticmethod
    def _fromDict(obj):
        """
        Returns a Event for the given dict.
        """
        typ = obj['_type']
        assert (typ in {"SystemEvent", "ObserveEvent"})
        paramSet = list(map(lambda p: Parameter._fromDict(p, True), obj['paramSet']))
        eventTime = EventTime._fromDict(obj['eventTime'])
        prefix = Prefix.from_str(obj['source'])
        eventName = EventName(obj['eventName'])
        eventId = obj['eventId']
        if typ == 'SystemEvent':
            return SystemEvent(prefix, eventName, paramSet, eventTime, eventId)
        else:
            return ObserveEvent(prefix, eventName, paramSet, eventTime, eventId)

    def _asDict(self):
        """
        Returns: dict
            a dictionary corresponding to this object
        """
        return {
            "_type": self.__class__.__name__,
            'eventId': self.eventId,
            'source': str(self.source),
            'eventName': self.eventName.name,
            'eventTime': self.eventTime._asDict(),
                'paramSet': list(map(lambda p: p._asDict(True), self.paramSet))
        }

    def isInvalid(self):
        return self.eventId == "-1"

    def get(self, keyName: str):
        """
        Gets the parameter with the given name, or else returns None

        Args:
            keyName (str): parameter name

        Returns: Parameter|None
            the parameter, if found
        """
        for p in self.paramSet:
            if p.keyName == keyName:
                return p

    def exists(self, keyName: str):
        """
        Returns true if the parameter with the given name is present in the event

        Args:
            keyName (str): parameter name

        Returns: bool
            true if the parameter is found
        """
        for p in self.paramSet:
            if p.keyName == keyName:
                return True
        return False


@dataclass
class SystemEvent(Event):
    """
    An event type for publishing changes in system data.

    Args:
        source (Prefix): prefix representing source of the event
        eventName (EventName): the name of event
        paramSet (list): list of Parameter (keys with values)
        eventTime (EventTime): the time the event was created (defaults to current time)
        eventId (str): event id (optional: Should leave empty unless received from event service)

    """
    @abstractmethod
    def eventType(self) -> str:
        return "SystemEvent"


@dataclass
class ObserveEvent(Event):
    """
    An event type fired when an observation has taken place.

    Args:
        source (Prefix): prefix representing source of the event
        eventName (EventName): the name of event
        paramSet (list): list of Parameter (keys with values)
        eventTime (EventTime): the time the event was created (defaults to current time)
        eventId (str): event id (optional: Should leave empty unless received from event service)

    """
    @abstractmethod
    def eventType(self) -> str:
        return "ObserveEvent"



import json
import traceback
from dataclasses import dataclass
from typing import List

import aiohttp
import structlog
from aiohttp import web, WSMessage
from aiohttp.web_request import Request
from aiohttp.web_response import Response
import atexit
import uuid

from aiohttp.web_ws import WebSocketResponse
from dataclasses_json import dataclass_json

from csw.CommandResponseManager import CommandResponseManager
from csw.ComponentHandlers import ComponentHandlers
from csw.ControlCommand import ControlCommand
from csw.Prefix import Prefix
from csw.LocationService import LocationService, ConnectionInfo, ComponentType, ConnectionType, HttpRegistration

# Ignore generated functions in API docs
__pdoc__ = {}

log = structlog.get_logger()


def _pdocIgnoreGenerated(className: str):
    __pdoc__[f"{className}.from_dict"] = False
    __pdoc__[f"{className}.from_json"] = False
    __pdoc__[f"{className}.schema"] = False
    __pdoc__[f"{className}.to_dict"] = False
    __pdoc__[f"{className}.to_json"] = False


_pdocIgnoreGenerated("Validate")


@dataclass_json
@dataclass
class Validate:
    """
    A message sent to validate a command. The response should be one of: Accepted, Invalid or Locked.

    Args:
        controlCommand (ControlCommand): The command to send
    """
    controlCommand: ControlCommand


_pdocIgnoreGenerated("Submit")


@dataclass_json
@dataclass
class Submit:
    """
    Represents a command that requires a response (of type CommandResponse).

    Args:
        controlCommand (ControlCommand): The command to send
    """
    controlCommand: ControlCommand


_pdocIgnoreGenerated("Oneway")


@dataclass_json
@dataclass
class Oneway:
    """
    Represents a command that does not require or expect a response

    Args:
        controlCommand (ControlCommand): The command to send
    """
    controlCommand: ControlCommand


_pdocIgnoreGenerated("QueryFinal")


@dataclass_json
@dataclass
class QueryFinal:
    """
    A message sent to query the final result of a long running command.
    The response should be a CommandResponse.

    Args:
        runId (str): The command's runId
        timeoutInSeconds (int) amount of time to wait
    """
    runId: str
    timeoutInSeconds: int

    @staticmethod
    def _fromDict(obj):
        """
        Returns a ControlCommand for the given dict.
        """
        try:
            runId = obj['runId']
            timeoutInSeconds = obj['timeoutInSeconds']
            return QueryFinal(runId, timeoutInSeconds)
        except:
            traceback.print_exc()


_pdocIgnoreGenerated("SubscribeCurrentState")


@dataclass_json
@dataclass
class SubscribeCurrentState:
    """
    Message used to subscribe to the current state of a component.

    Args:
        stateNames (List[str]) list of current state names to subscribe to
    """
    stateNames: List[str]

    @staticmethod
    def _fromDict(obj):
        """
        Returns a SubscribeCurrentState for the given dict.
        """
        # typ = obj["_type"]
        if "names" in obj.keys():
            stateNames = obj.get("names")
        else:
            stateNames = []
        return SubscribeCurrentState(stateNames)


class CommandServer:
    _app = web.Application()
    _crm = CommandResponseManager()

    async def _handlePost(self, request: Request) -> Response:
        obj = await request.json()
        # log.debug(f"received post: {str(obj)}")
        method = obj['_type']
        if method in {'Submit', 'Oneway', 'Validate'}:
            command = ControlCommand._fromDict(obj['controlCommand'])
            runId = str(uuid.uuid4())
            if method == 'Submit':
                commandResponse, task = self.handler.onSubmit(runId, command)
                if task is not None:
                    # noinspection PyTypeChecker
                    self._crm.addTask(runId, task)
                    log.debug("Long running task in progress...")
            elif method == 'Oneway':
                commandResponse = self.handler.onOneway(runId, command)
            else:
                commandResponse = self.handler.validateCommand(runId, command)
            responseDict = commandResponse._asDict()
            return web.json_response(responseDict)
        else:
            raise Exception("Invalid command type: " + method)

    async def _handleQueryFinal(self, queryFinal: QueryFinal) -> Response:
        commandResponse = await self._crm.waitForTask(queryFinal.runId, queryFinal.timeoutInSeconds)
        responseDict = commandResponse._asDict()
        return web.json_response(responseDict)

    async def _handleWs(self, request: Request) -> WebSocketResponse:
        ws = web.WebSocketResponse()
        await ws.prepare(request)
        msg: WSMessage
        async for msg in ws:
            if msg.type == aiohttp.WSMsgType.TEXT:
                if msg.data == 'close':
                    log.debug("Received ws close message")
                    await ws.close()
                else:
                    obj = json.loads(msg.data)
                    # log.debug(f"received message: {str(obj)}")
                    method = obj['_type']
                    if method == "QueryFinal":
                        queryFinal = QueryFinal._fromDict(obj)
                        resp = await self._handleQueryFinal(queryFinal)
                        await ws.send_str(resp.text)
                        await ws.close()
                    elif method == "SubscribeCurrentState":
                        stateNames = SubscribeCurrentState._fromDict(obj).stateNames
                        log.debug(f"Received SubscribeCurrentState: stateNames = {stateNames}")
                        self.handler._subscribeCurrentState(stateNames, ws)
                    else:
                        log.debug(f"Warning: Received unknown ws message: {str(msg.data)}")
            elif msg.type == aiohttp.WSMsgType.ERROR:
                log.debug('Error: ws connection closed with exception %s' % ws.exception())
        log.debug('websocket connection closed')
        self.handler._unsubscribeCurrentState(ws)
        return ws

    @staticmethod
    def _registerWithLocationService(prefix: Prefix, port: int):
        log.debug("Registering with location service using port " + str(port))
        locationService = LocationService()
        connection = ConnectionInfo.make(prefix, ComponentType.Service, ConnectionType.HttpType)
        atexit.register(locationService.unregister, connection)
        # locationService.unregister(connection)
        locationService.register(HttpRegistration(connection, port, "/post-endpoint"))

    def __init__(self, prefix: Prefix, handler: ComponentHandlers, port: int = 8082):
        """
        Creates an HTTP server that can receive CSW commands and registers it with the Location Service using the given prefix,
        so that CSW components can locate it and send commands to it.

        Args:
            prefix (str): a CSW Prefix in the format $subsystem.name, where subsystem is one of the upper case TMT
                          subsystem names and name is the name of the command server
            handler (ComponentHandlers): command handler notified when commands are received
            port (int): optional port for HTTP server
        """
        self.handler = handler
        self.port = port
        self._app.add_routes([
            web.post('/post-endpoint', self._handlePost),
            web.get("/websocket-endpoint", self._handleWs)
        ])
        self._registerWithLocationService(prefix, port)

    def start(self):
        """
        Starts the command http server in a thread
        """
        web.run_app(self._app, port=self.port)



import traceback
from dataclasses import dataclass
from typing import List
from csw.Parameter import Parameter
from csw.Prefix import Prefix


@dataclass
class CurrentState:
    """
    Represents the current state of a python based CSW component.
    """
    prefix: Prefix
    stateName: str
    paramSet: List[Parameter]

    @staticmethod
    def _fromDict(obj):
        """
        Returns a CurrentState for the given dict.
        """
        prefix = Prefix.from_str(obj['prefix'])
        stateName = obj['stateName']
        paramSet = list(map(lambda p: Parameter._fromDict(p), obj['paramSet']))
        return CurrentState(prefix, stateName, paramSet)

    def _asDict(self):
        """
        Returns: dict
            a dictionary corresponding to this object
        """
        try:
            d = {
                'prefix': str(self.prefix),
                'stateName': self.stateName,
                'paramSet': list(map(lambda p: p._asDict(), self.paramSet))
            }
            return d
        except Exception as ex:
            traceback.print_exc()

    def get(self, keyName: str):
        """
        Gets the parameter with the given name, or else returns None

        Args:
            keyName (str): parameter name

        Returns: Parameter|None
            the parameter, if found
        """
        for p in self.paramSet:
            if p.keyName == keyName:
                return p

    def exists(self, keyName: str):
        """
        Returns true if the parameter with the given name is present

        Args:
            keyName (str): parameter name

        Returns: bool
            true if the parameter is found
        """
        for p in self.paramSet:
            if p.keyName == keyName:
                return True
        return False



import cbor2

from csw.RedisConnector import RedisConnector
from csw.Event import Event
from csw.EventKey import EventKey

class EventSubscriber:

    def __init__(self):
        self.__redis = RedisConnector()

    @staticmethod
    def __handleCallback(message: dict, callback):
        data = message['data']
        event = Event._fromDict(cbor2.loads(data))
        callback(event)

    def subscribe(self, eventKeyList: list[EventKey], callback):
        """
        Start a subscription to system events in event service, specifying a callback
        to be called when an event in the list has its value updated.

        Args:
            eventKeyList (list[EventKey]): list of event EventKey to subscribe to
            callback (function): function to be called when event updates. Should take Event and return void

        Returns: PubSubWorkerThread
            subscription thread. Use .stop() method to stop subscription.
        """
        keyList = list(map(lambda k: str(k), eventKeyList))
        return self.__redis.subscribe(keyList, lambda message: self.__handleCallback(message, callback))

    def unsubscribe(self, eventKeyList: list[EventKey]):
        """
        Unsubscribes to the given list of event keys (or all keys, if eventKeyList is empty)

        Args:
            eventKeyList (list[EventKey]): list of EventKeys to unsubscribe from
        """
        keyList = list(map(lambda k: str(k), eventKeyList))
        return self.__redis.unsubscribe(keyList)

    # XXX Commented out due to Event Service performance concerns when using psubscribe
    # def pSubscribe(self, eventKeyList: list, callback):
    #     """
    #     Start a subscription to system events in event service, specifying a callback
    #     to be called when an event in the list has its value updated.
    #     In this case the keys are treated as glob-style patterns:
    #     h?llo subscribes to hello, hallo and hxllo,
    #     h*llo subscribes to hllo and heeeello,
    #     h[ae]llo subscribes to hello and hallo, but not hillo.
    #
    #     Args:
    #         eventKeyList (list): list of event key (string patterns) to subscribe to
    #         callback (function): function to be called when event updates. Should take Event and return void
    #
    #     Returns: PubSubWorkerThread
    #         subscription thread. Use .stop() method to stop subscription.
    #     """
    #     return self.__redis.pSubscribe(eventKeyList, lambda message: self.__handleCallback(message, callback))

    # def pUnsubscribe(self, eventKeyList: list):
    #     """
    #     Unsubscribes to the given list of event key patterns (or all keys, if eventKeyList is empty)
    #
    #     Args:
    #         eventKeyList (list): list of event key patterns (Strings) to unsubscribe from
    #     """
    #     return self.__redis.pUnsubscribe(eventKeyList)

    def get(self, eventKey: EventKey):
        """
        Get an event from the Event Service

        Args:
            eventKey (EventKey): String specifying Redis key for event.  Should be source prefix + "." + event name.

        Returns: Event
            Event obtained from Event Service, decoded into a Event
        """
        data = self.__redis.get(str(eventKey))
        event = Event._fromDict(cbor2.loads(data))
        return event



from dataclasses import dataclass, asdict
from enum import Enum
from astropy.coordinates import Angle
from astropy import units as u


@dataclass
class Tag:
    """
    A tag is a label to indicate the use of the coordinate
    """
    name: str


# noinspection PyUnresolvedReferences
@dataclass
class _CswAngle:
    """
    A wrapper for Angle values (See the csw.params.core.models.Angle class).
    Note: The Python API uses astropy Angle. This class is used internally for compatibility.
    Normally Angle would be stored in double as radians, but this introduces rounding errors.
    This class stores value in microarc seconds to prevent rounding errors.
    """
    uas: int

    def toAngle(self):
        return Angle(self.uas * u.uarcsec)

    @staticmethod
    def fromAngle(a: Angle):
        return _CswAngle(int(a.uarcsec))


class EqFrame(Enum):
    ICRS = 0
    FK5 = 1


class SolarSystemObject(Enum):
    """
    Enum type: Represents the possible values for a Solar System Object.
    """
    Mercury = 0
    Venus = 1
    Moon = 2
    Mars = 3
    Jupiter = 4
    Saturn = 5
    Neptune = 6
    Uranus = 7
    Pluto = 8


@dataclass
class ProperMotion:
    pmx: float
    pmy: float


@dataclass
class Coord:
    """
    This is the base class of the coordinate types.
    If the key type is CoordKey, the value type can be any of the Coord subtypes.
    The dict key gives the class name.

    Args:
        tag (Tag): a label to indicate the use of the coordinate

    """
    tag: Tag

    @staticmethod
    def _fromDict(obj: dict):
        switcher = {
            "EqCoord": EqCoord,
            "SolarSystemCoord": SolarSystemCoord,
            "MinorPlanetCoord": MinorPlanetCoord,
            "CometCoord": CometCoord,
            "AltAzCoord": AltAzCoord
        }
        typ = obj["_type"]
        return switcher[typ]._fromValueDict(obj)

    def _asDict(self):
        pass


# noinspection PyUnresolvedReferences
@dataclass
class EqCoord(Coord):
    """
    Represents equatorial coordinates (mirrors class of same name in the CSW Scala code).
    """
    ra: Angle
    dec: Angle
    frame: EqFrame
    catalogName: str
    pm: ProperMotion

    @staticmethod
    def make(tag: str = "BASE", ra: any = "0 deg", dec: any = "0 deg", frame: EqFrame = EqFrame.ICRS,
             catalogName: str = "none", pm: tuple = (0.0, 0.0)):
        """
        Convenience factory method.
        Note that the ra and dec args should be in a format accepted by astropy's Angle class.
        """
        return EqCoord(Tag(tag), Angle(ra), Angle(dec), frame, catalogName, ProperMotion(pm[0], pm[1]))

    def _asDict(self):
        return {
            "_type": self.__class__.__name__,
            "tag": self.tag.name,
            "ra": _CswAngle.fromAngle(self.ra).uas,
            "dec": _CswAngle.fromAngle(self.dec).uas,
            "frame": self.frame.name,
            "catalogName": self.catalogName,
            "pm": asdict(self.pm)
        }

    @staticmethod
    def _fromValueDict(obj: dict):
        return EqCoord(
            tag=Tag(obj["tag"]),
            ra=_CswAngle(obj["ra"]).toAngle().to(u.deg),
            dec=_CswAngle(obj["dec"]).toAngle().to(u.deg),
            frame=EqFrame[obj["frame"]],
            catalogName=obj["catalogName"],
            pm=ProperMotion(**obj["pm"]),
        )


@dataclass
class SolarSystemCoord(Coord):
    """
    Represents Solar System Coordinates (mirrors class of same name in the CSW Scala code).
    """
    body: SolarSystemObject

    @staticmethod
    def make(tag: str, body: SolarSystemObject):
        return SolarSystemCoord(Tag(tag), body)

    def _asDict(self):
        return {
            "_type": self.__class__.__name__,
            "tag": self.tag.name,
            "body": self.body.name,
        }

    @staticmethod
    def _fromValueDict(obj: dict):
        return SolarSystemCoord(
            tag=Tag(obj["tag"]),
            body=SolarSystemObject[obj["body"]]
        )


# noinspection PyUnresolvedReferences
@dataclass
class MinorPlanetCoord(Coord):
    """
    Represents Minor Planet Coordinates (mirrors class of same name in the CSW Scala code).
    """
    epoch: float  # TT as a Modified Julian Date
    inclination: Angle
    longAscendingNode: Angle
    argOfPerihelion: Angle
    meanDistance: float  # AU
    eccentricity: float
    meanAnomaly: Angle

    @staticmethod
    def make(tag: str, epoch: float, inclination: any, longAscendingNode: any, argOfPerihelion: any,
             meanDistance: float, eccentricity: float, meanAnomaly: any):
        """
        Convenience factory method.
        Note that the inclination, longAscendingNode, argOfPerihelion and meanAnomaly args
        should be in a format accepted by astropy's Angle class.
        """
        return MinorPlanetCoord(Tag(tag), epoch, Angle(inclination), Angle(longAscendingNode), Angle(argOfPerihelion),
                                meanDistance, eccentricity, Angle(meanAnomaly))

    def _asDict(self):
        return {
            "_type": self.__class__.__name__,
            "tag": self.tag.name,
            "epoch": self.epoch,
            "inclination": _CswAngle.fromAngle(self.inclination).uas,
            "longAscendingNode": _CswAngle.fromAngle(self.longAscendingNode).uas,
            "argOfPerihelion": _CswAngle.fromAngle(self.argOfPerihelion).uas,
            "meanDistance": self.meanDistance,
            "eccentricity": self.eccentricity,
            "meanAnomaly": _CswAngle.fromAngle(self.meanAnomaly).uas
        }

    @staticmethod
    def _fromValueDict(obj: dict):
        return MinorPlanetCoord(
            tag=Tag(obj["tag"]),
            epoch=obj["epoch"],
            inclination=_CswAngle(obj["inclination"]).toAngle().to(u.deg),
            longAscendingNode=_CswAngle(obj["longAscendingNode"]).toAngle().to(u.deg),
            argOfPerihelion=_CswAngle(obj["argOfPerihelion"]).toAngle().to(u.deg),
            meanDistance=obj["meanDistance"],
            eccentricity=obj["eccentricity"],
            meanAnomaly=_CswAngle(obj["meanAnomaly"]).toAngle().to(u.deg)
        )


# noinspection PyUnresolvedReferences
@dataclass
class CometCoord(Coord):
    """
    Represents Comet Coordinates (mirrors class of same name in the CSW Scala code).
    """
    epochOfPerihelion: float  # TT as a Modified Julian Date
    inclination: Angle
    longAscendingNode: Angle
    argOfPerihelion: Angle
    perihelionDistance: float  # AU
    eccentricity: float

    @staticmethod
    def make(tag: str, epochOfPerihelion: float, inclination: any, longAscendingNode: any, argOfPerihelion: any,
             perihelionDistance: float, eccentricity: float):
        """
        Convenience factory method.
        Note that the inclination, longAscendingNode and argOfPerihelion args
        should be in a format accepted by astropy's Angle class.
        """
        return CometCoord(Tag(tag), epochOfPerihelion, Angle(inclination), Angle(longAscendingNode),
                          Angle(argOfPerihelion),
                          perihelionDistance, eccentricity)

    def _asDict(self):
        return {
            "_type": self.__class__.__name__,
            "tag": self.tag.name,
            "epochOfPerihelion": self.epochOfPerihelion,
            "inclination": _CswAngle.fromAngle(self.inclination).uas,
            "longAscendingNode": _CswAngle.fromAngle(self.longAscendingNode).uas,
            "argOfPerihelion": _CswAngle.fromAngle(self.argOfPerihelion).uas,
            "perihelionDistance": self.perihelionDistance,
            "eccentricity": self.eccentricity
        }

    @staticmethod
    def _fromValueDict(obj: dict):
        return CometCoord(
            tag=Tag(obj["tag"]),
            epochOfPerihelion=obj["epochOfPerihelion"],
            inclination=_CswAngle(obj["inclination"]).toAngle().to(u.deg),
            longAscendingNode=_CswAngle(obj["longAscendingNode"]).toAngle().to(u.deg),
            argOfPerihelion=_CswAngle(obj["argOfPerihelion"]).toAngle().to(u.deg),
            perihelionDistance=obj["perihelionDistance"],
            eccentricity=obj["eccentricity"]
        )


# noinspection PyUnresolvedReferences
@dataclass
class AltAzCoord(Coord):
    """
    Represents Alt-Az Coordinates (mirrors class of same name in the CSW Scala code).
    """
    alt: Angle = Angle("0 deg")
    az: Angle = Angle("0 deg")

    @staticmethod
    def make(tag: str = "BASE", alt: any = "0 deg", az: any = "0 deg"):
        """
        Convenience factory method.
        Note that the alt and az args should be in a format accepted by astropy's Angle class.
        """
        return AltAzCoord(Tag(tag), Angle(alt), Angle(az))

    def _asDict(self):
        return {
            "_type": self.__class__.__name__,
            "tag": self.tag.name,
            "alt": _CswAngle.fromAngle(self.alt).uas,
            "az": _CswAngle.fromAngle(self.az).uas
        }

    @staticmethod
    def _fromValueDict(obj: dict):
        return AltAzCoord(
            tag=Tag(obj["tag"]),
            alt=_CswAngle(obj["alt"]).toAngle().to(u.deg),
            az=_CswAngle(obj["az"]).toAngle().to(u.deg),
        )



from typing import List
from dataclasses import dataclass, field

import structlog
from dataclasses_json import dataclass_json
from enum import Enum
from multipledispatch import dispatch
import requests
import json
from csw.Prefix import Prefix

# Ignore generated functions in API docs
# noinspection SpellCheckingInspection
__pdoc__ = {}

# Note: Update this before each release!
def csw_version():
    return {"csw-version": "4.0.1"}

# noinspection SpellCheckingInspection
def _pdocIgnoreGenerated(className: str):
    __pdoc__[f"{className}.from_dict"] = False
    __pdoc__[f"{className}.from_json"] = False
    __pdoc__[f"{className}.schema"] = False
    __pdoc__[f"{className}.to_dict"] = False
    __pdoc__[f"{className}.to_json"] = False


# Python API for CSW Location Service

# TODO: Add Location Service subscribe / TrackingEvent

#  "ComponentType" : [ "Container", "HCD", "Assembly", "Sequencer", "SequenceComponent", "Service", "Machine" ],

class ComponentType(Enum):
    """
    Enum type: Represents a type of the CSW Component.
    """
    Assembly = "Assembly"
    HCD = "HCD"
    Container = "Container"
    Service = "Service"
    Sequencer = "Sequencer"
    SequenceComponent = "SequenceComponent"
    Machine = "Machine"


class ConnectionType(Enum):
    """
    Enum type: Represents a type of connection offered by the Component.
    Note that AkkaType is for Akka Actor based connections.
    Python applications can only communicate via HTTP or TCP connections.
    """
    HttpType = "http"
    TcpType = "tcp"
    AkkaType = "akka"


_pdocIgnoreGenerated("ConnectionInfo")


@dataclass_json
@dataclass
class ComponentId:
    prefix: str
    componentType: str


@dataclass_json
@dataclass
class ConnectionInfo:
    prefix: str
    componentType: str
    connectionType: str

    @staticmethod
    def make(prefix: Prefix, componentType: ComponentType, connectionType: ConnectionType):
        return ConnectionInfo(str(prefix), componentType.value, connectionType.value)


_pdocIgnoreGenerated("ResolveInfo")


@dataclass_json
@dataclass
class ResolveInfo:
    _type: str
    connection: ConnectionInfo
    within: str


_pdocIgnoreGenerated("Location")


@dataclass_json
@dataclass
class Location:
    _type: str
    connection: ConnectionInfo
    uri: str
    metadata: dict

    @staticmethod
    def _makeLocation(obj: dict):
        typ = obj["_type"]
        s = json.dumps(obj)
        if typ == "HttpLocation":
            return HttpLocation.schema().loads(s)
        elif typ == "TcpLocation":
            return TcpLocation.schema().loads(s)
        elif typ == "AkkaLocation":
            return AkkaLocation.schema().loads(s)
        else:
            raise Exception("Invalid location type: " + typ)


_pdocIgnoreGenerated("AkkaLocation")


@dataclass_json
@dataclass
class AkkaLocation(Location):
    pass


_pdocIgnoreGenerated("TcpLocation")


@dataclass_json
@dataclass
class TcpLocation(Location):
    pass


_pdocIgnoreGenerated("HttpLocation")


@dataclass_json
@dataclass
class HttpLocation(Location):
    pass


_pdocIgnoreGenerated("Registration")


@dataclass_json
@dataclass
class Registration:
    """
    Abstract base class for registering with the location service
    """
    connection: ConnectionInfo


_pdocIgnoreGenerated("NetworkType")


@dataclass_json
@dataclass
class NetworkType:
    """
    NetworkType enum {Outside, Inside)
    """
    _type: str


_pdocIgnoreGenerated("HttpRegistration")


@dataclass_json
@dataclass
class HttpRegistration(Registration):
    """
    Used to register an http based service with the Location Service.
    """
    port: int
    path: str = ""
    networkType: NetworkType = NetworkType("Inside")
    metadata: dict = field(default_factory=csw_version)
    _type: str = "HttpRegistration"


@dataclass_json
@dataclass
class AkkaRegistration(Registration):
    """
    Used to register an http based service with the Location Service.
    """
    actorRefURI: str
    metadata: dict = field(default_factory=dict)
    _type: str = "AkkaRegistration"


_pdocIgnoreGenerated("TcpRegistration")


@dataclass_json
@dataclass
class TcpRegistration(Registration):
    """
    Used to register a tcp based service with the Location Service.
    """
    port: int
    metadata: dict = field(default_factory=dict)
    _type: str = "TcpRegistration"


# noinspection PyProtectedMember
class LocationService:
    log = structlog.get_logger()
    baseUri = "http://127.0.0.1:7654/"
    postUri = f"{baseUri}post-endpoint"
    wsUri = f"{baseUri}websocket-endpoint"

    def register(self, registration: Registration) -> ConnectionInfo:
        """
        Registers a connection.

        Args:
            registration (Registration): the Registration holding the connection and it's corresponding location to
            register with LocationService

        Returns: ConnectionInfo
            an object describing the connection
        """
        # regType = registration.__class__.__name__
        regJson = json.loads(registration.to_json())
        regJson['_type'] = registration.__class__.__name__
        jsonBody = f'{{"_type": "Register", "registration": {json.dumps(regJson)}}}'
        r = requests.post(self.postUri, json=json.loads(jsonBody))
        if not r.ok:
            raise Exception(r.text)
        return registration.connection

    def unregister(self, connection: ConnectionInfo):
        """
        Unregisters a connection.

        Args:
            connection (ConnectionInfo): an already registered connection
        """
        self.log.debug(f"Unregistering connection {connection} from the Location Service.")
        jsonBody = f'{{"_type": "Unregister", "connection": {connection.to_json()}}}'
        r = requests.post(self.postUri, json=json.loads(jsonBody))
        if not r.ok:
            raise Exception(r.text)

    def _postJson(self, jsonBody: str) -> Location:
        r = requests.post(self.postUri, json=json.loads(jsonBody))
        if not r.ok:
            raise Exception(r.text)
        maybeResult = json.loads(r.text)
        if len(maybeResult) != 0:
            return Location._makeLocation(maybeResult[0])

    def find(self, connection: ConnectionInfo) -> Location:
        """
        Resolves the location for a connection from the local cache

        Args:
            connection (ConnectionInfo): an already registered connection

        Returns: Location
            the Location
        """
        return self._postJson(f'{{"_type": "Find", "connection": {connection.to_json()}}}')

    # "within":"2 seconds"}}
    def resolve(self, connection: ConnectionInfo, withinSecs: int = "5") -> Location:
        """
        Resolves the location for a connection from the local cache, if not found waits for the event to arrive
        within specified time limit

        Args:
            connection (ConnectionInfo): an already registered connection
            withinSecs (int): max number of seconds to wait for the connection to be found

        Returns: Location
            the Location
        """
        resolveInfo = ResolveInfo("Resolve", connection, f"{withinSecs} seconds")
        return self._postJson(resolveInfo.to_json())

    @staticmethod
    def _list(jsonBody: str) -> List[Location]:
        r = requests.post(LocationService.postUri, json=json.loads(jsonBody))
        if not r.ok:
            raise Exception(r.text)
        return list(map(lambda x: Location._makeLocation(x), json.loads(r.text)))

    @dispatch()
    def list(self) -> List[Location]:
        """
        Lists all locations registered.

        Returns: List[Location]
            list of locations
        """
        jsonBody = '{"_type": "ListEntries"}'
        return self._list(jsonBody)

    @dispatch(ComponentType)
    def list(self, componentType: ComponentType) -> List[Location]:
        """
        Lists components of the given component type

        Args:
            componentType (ComponentType): the type of component to list

        Returns: List[Location]
            list of locations
        """
        jsonBody = f'{{"_type": "ListByComponentType", "componentType": "{componentType.value}"}}'
        return self._list(jsonBody)

    @dispatch(str)
    def list(self, hostname: str) -> List[Location]:
        """
        Lists all locations registered on the given hostname

        Args:
            hostname (str): restrict result to this host

        Returns: List[Location]
            list of locations
        """
        jsonBody = f'{{"_type": "ListByHostname", "hostname": "{hostname}"}}'
        return self._list(jsonBody)

    @dispatch(ConnectionType)
    def list(self, connectionType: ConnectionType) -> List[Location]:
        """
        Lists all locations registered with the given connection type

        Args:
            connectionType (ConnectionType): restrict result to this connection type

        Returns: List[Location]
            list of locations
        """
        jsonBody = f'{{"_type": "ListByConnectionType", "connectionType": "{connectionType.value}"}}'
        return self._list(jsonBody)

    def listByPrefix(self, prefix: Prefix) -> List[Location]:
        """
        Lists all locations with the given prefix

        Args:
            prefix (Prefix): restrict result to this prefix

        Returns: List[Location]
            list of locations
        """
        jsonBody = f'{{"_type": "ListByPrefix", "prefix": "{str(prefix)}"}}'
        return self._list(jsonBody)



from datetime import datetime,timezone
from dataclasses import dataclass, asdict


@dataclass
class EventTime:
    """
    Creates an EventTime containing seconds since the epoch (1970) and the offset from seconds in nanoseconds
    """
    seconds: int
    nanos: int

    def __str__(self):
        secs = self.seconds + self.nanos / 1e9
        dt = datetime.fromtimestamp(secs, timezone.utc)
        return dt.strftime('%Y-%m-%dT%H:%M:%S.%f')

    def _asDict(self):
        return asdict(self)

    @staticmethod
    def _fromDict(obj: dict):
        return EventTime(**obj)

    @staticmethod
    def fromSystem():
        """
        Returns an EventTime with the current time.
        """
        t = datetime.now(timezone.utc).timestamp()
        seconds = int(t)
        nanos = int((t - seconds) * 1e9)
        return EventTime(seconds, nanos)



from dataclasses import dataclass
from typing import List
from csw.Parameter import Parameter
from csw.Prefix import Prefix

@dataclass
class CommandName:
    """
    A wrapper class representing the name of a Command
    """
    name: str

@dataclass
class ControlCommand:
    """
    Represents a CSW command.
    """
    source: Prefix
    commandName: CommandName
    maybeObsId: List[str]
    paramSet: List[Parameter]

    # noinspection PyProtectedMember
    @staticmethod
    def _fromDict(obj):
        """
        Returns a ControlCommand for the given dict.
        """
        typ = obj["_type"]
        source = Prefix.from_str(obj['source'])
        commandName = CommandName(obj['commandName'])
        maybeObsId = obj['maybeObsId'] if 'maybeObsId' in obj else ""
        paramSet = list(map(lambda p: Parameter._fromDict(p), obj['paramSet']))
        assert (typ in {"Setup", "Observe"})
        if typ == 'Setup':
            return Setup(source, commandName, maybeObsId, paramSet)
        else:
            return Observe(source, commandName, maybeObsId, paramSet)

    # noinspection PyProtectedMember
    def _asDict(self):
        """
        Returns: dict
            a dictionary corresponding to this object
        """
        d = {
            '_type': self.__class__.__name__,
            'source': str(self.source),
            'commandName': self.commandName.name,
            'paramSet': list(map(lambda p: p._asDict(), self.paramSet))
        }
        if len(self.maybeObsId) != 0:
            d['maybeObsId'] = self.maybeObsId

        return d

    def get(self, keyName: str):
        """
        Gets the parameter with the given name, or else returns None.

        Args:
            keyName (str): parameter name

        Returns: Parameter|None
            the parameter, if found
        """
        for p in self.paramSet:
            if p.keyName == keyName:
                return p

    def exists(self, keyName: str):
        """
        Returns true if the parameter with the given name is present

        Args:
            keyName (str): parameter name

        Returns: bool
            true if the parameter is found
        """
        for p in self.paramSet:
            if p.keyName == keyName:
                return True
        return False


@dataclass
class Setup(ControlCommand):
    """
    A Setup is a command that can be sent to an HCD or Assembly.
    """
    pass


@dataclass
class Observe(ControlCommand):
    """
    An Observe is a special command that can be sent to an HCD or Assembly.
    """
    pass



from datetime import datetime, timezone
from dataclasses import dataclass, asdict
from dateutil import parser

@dataclass
class UTCTime:
    """
    Creates a UTCTime containing seconds since the epoch (1970) and the offset from seconds in nanoseconds
    """
    seconds: int
    nanos: int

    def __str__(self):
        secs = self.seconds + self.nanos / 1e9
        dt = datetime.fromtimestamp(secs, timezone.utc)
        return dt.strftime('%Y-%m-%dT%H:%M:%S.%f') + "Z"

    def _asDict(self):
        return asdict(self)

    @staticmethod
    def _fromDict(obj: dict):
        return UTCTime(**obj)

    @staticmethod
    def from_str(timeStr: str):
        """
        Returns a UTCTime given a string in ISO format (ex: "2021-09-20T18:44:12.419084072Z").
        """
        t = parser.isoparse(timeStr).timestamp()
        seconds = int(t)
        nanos = int((t - seconds) * 1e9)
        return UTCTime(seconds, nanos)

    @staticmethod
    def fromSystem():
        """
        Returns a UTCTime with the current time.
        """
        t = datetime.now(timezone.utc).timestamp()
        seconds = int(t)
        nanos = int((t - seconds) * 1e9)
        return UTCTime(seconds, nanos)



from datetime import datetime, timezone
from dataclasses import dataclass, asdict
from astropy.time import Time
from dateutil import parser


@dataclass
class TAITime:
    """
    Creates a TAITime containing seconds since the epoch (1970) and the offset from seconds in nanoseconds
    """
    seconds: int
    nanos: int

    def __str__(self):
        secs = self.seconds + self.nanos / 1e9
        dt = datetime.fromtimestamp(secs, timezone.utc)
        return dt.strftime('%Y-%m-%dT%H:%M:%S.%f') + "Z"

    def _asDict(self):
        return asdict(self)

    @staticmethod
    def _fromDict(obj: dict):
        return TAITime(**obj)

    @staticmethod
    def from_str(timeStr: str):
        """
        Returns a TAITime given a string in ISO format (ex: "2021-09-20T18:44:12.419084072Z").
        """
        t = parser.isoparse(timeStr).timestamp()
        seconds = int(t)
        nanos = int((t - seconds) * 1e9)
        return TAITime(seconds, nanos)

    @staticmethod
    def fromSystem():
        """
        Returns a TAITime with the current time.
        """
        t = Time.now().tai.value.timestamp()
        seconds = int(t)
        nanos = int((t - seconds) * 1e9)
        return TAITime(seconds, nanos)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="tmtpycsw",
    version="4.0.2",
    python_requires='>=3.9.5',
    author="Allan Brighton",
    author_email="allanexus@gmail.com",
    description="A python API for TMT CSW services",
    long_description="This package provides some Python APIs for TMT Common Software (CSW) services.",
    long_description_content_type="text/markdown",
    url="https://github.com/tmtsoftware/pycsw",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from dataclasses import dataclass

from csw.Prefix import Prefix
from csw.Event import EventName

@dataclass
class EventKey:
    """
    A wrapper class representing the key for an event

    Args:
        source (Prefix): represents the prefix of the component that publishes this event
        eventName (EventName): represents the name of the event
    """
    source: Prefix
    eventName: EventName

    def __str__(self):
        return f"{str(self.source)}.{self.eventName.name}"

    @classmethod
    def from_str(class_object, eventKeyStr: str):
        i = eventKeyStr.rindex('.')
        return EventKey(Prefix.from_str(eventKeyStr[:i]), EventName(eventKeyStr[i+1:]))



import json
from typing import List

import structlog
from aiohttp.web_ws import WebSocketResponse

from csw.CommandResponse import CommandResponse, Accepted, Error
from csw.ControlCommand import ControlCommand
from asyncio import Task

from csw.CurrentState import CurrentState


class ComponentHandlers:
    """
    Abstract base class for handling CSW commands.
    Subclasses can override onSubmit, onOneway, validateCommand, and currentState to implement the behavior of the
    component.
    """

    log = structlog.get_logger()

    # map of current state name to list of websockets going to subscribers
    _currentStateSubscribers = {}

    def onSubmit(self, runId: str, command: ControlCommand) -> (CommandResponse, Task):
        """
        Handles the given setup command and returns a CommandResponse subclass

        Args:
            runId (str): unique id for this command
            command (ControlCommand): contains the command

        Returns: (CommandResponse, Task)
            a pair: (subclass of CommandResponse, Task), where the task can be None if the command response is final.
            For long running commands, you can respond with Started(runId, "...") and a task that completes the work in the background.
        """
        # noinspection PyTypeChecker
        return Error(runId, "Not implemented: submit command handler"), None

    def onOneway(self, runId: str, command: ControlCommand) -> CommandResponse:
        """
        Handles the given setup command and returns an immediate CommandResponse

        Args:
            runId (str): unique id for this command
            command (ControlCommand): contains the command

        Returns: CommandResponse
            a subclass of CommandResponse (only Accepted, Invalid or Locked are allowed)
        """
        return Error(runId, "Not implemented: oneway command handler")

    def validateCommand(self, runId: str, command: ControlCommand) -> CommandResponse:
        """
        Validates the given command

        Args:
            runId (str): unique id for this command
            command (ControlCommand): contains the command

        Returns: CommandResponse
            a subclass of CommandResponse (only Accepted, Invalid or Locked are allowed)
       """
        return Accepted(runId)

    def currentStates(self) -> List[CurrentState]:
        """
        Returns the current states for the component
        """
        return []

    # ---Do not override the following methods ---

    async def publishCurrentStates(self):
        """
        Publish the current state of the python based CSW component
        """
        for currentState in self.currentStates():
            s1 = self._currentStateSubscribers.get("") or set()
            s2 = self._currentStateSubscribers.get(currentState.stateName) or set()
            subscribers = s1 | s2
            dict = currentState._asDict()
            cs = json.dumps(dict)
            for ws in subscribers:
                self.log.debug(f"Publishing current state: {cs}")
                await ws.send_str(cs)

    def _addCurrentStateSubscriber(self, stateName: str, ws: WebSocketResponse):
        if (stateName in self._currentStateSubscribers):
            self._currentStateSubscribers[stateName].add(ws)
        else:
            self._currentStateSubscribers[stateName] = {ws}

    def _subscribeCurrentState(self, stateNames: List[str], ws: WebSocketResponse):
        """
        Internal method used to subscribe to current state of this component.
        If stateNames is empty, subscribe to all current states
        """
        if len(stateNames) == 0:
            self._addCurrentStateSubscriber("", ws)
        else:
            for stateName in stateNames:
                self._addCurrentStateSubscriber(stateName, ws)

    def _unsubscribeCurrentState(self, ws: WebSocketResponse):
        """
        Internal method used to unsubscribe a websocket from current state events
        """
        for stateName in self._currentStateSubscribers:
            if ws in self._currentStateSubscribers[stateName]:
                self._currentStateSubscribers[stateName].remove(ws)
